from pdfplumber import utils
from operator import itemgetter
import itertools

DEFAULT_SNAP_TOLERANCE = 3
DEFAULT_JOIN_TOLERANCE = 3
DEFAULT_MIN_WORDS_VERTICAL = 3
DEFAULT_MIN_WORDS_HORIZONTAL = 1

def move_to_avg(objs, orientation):
    """
    Move `objs` vertically/horizontally to their average x/y position.
    """
    if orientation not in ("h", "v"):
        raise ValueError("Orientation must be 'v' or 'h'")
    if len(objs) == 0: return []
    move_axis = "v" if orientation == "h" else "h"
    attr = "top" if orientation == "h" else "x0"
    values = list(map(itemgetter(attr), objs))
    q = pow(10, utils.decimalize(values[0]).as_tuple().exponent)
    avg = utils.decimalize(float(sum(values) / len(values)), q)
    new_objs = [ utils.move_object(obj, move_axis, avg - obj[attr])
        for obj in objs ]
    return new_objs

def snap_edges(edges, tolerance=DEFAULT_SNAP_TOLERANCE):
    """
    Given a list of edges, snap any within `tolerance` pixels of one another to their positional average.
    """
    v, h = [ list(filter(lambda x: x["orientation"] == o, edges))
        for o in ("v", "h") ]

    v = [ move_to_avg(cluster, "v")
        for cluster in utils.cluster_objects(v, "x0", tolerance) ]

    h = [ move_to_avg(cluster, "h")
        for cluster in utils.cluster_objects(h, "top", tolerance) ]

    snapped = list(itertools.chain(*(v + h)))
    return snapped

def join_edge_group(edges, orientation, tolerance=DEFAULT_JOIN_TOLERANCE):
    """
    Given a list of edges along the same infinite line, join those that are within `tolerance` pixels of one another.
    """
    if orientation == "h":
        min_prop, max_prop = "x0", "x1"
    elif orientation == "v":
        min_prop, max_prop = "top", "bottom"
    else:
        raise ValueError("Orientation must be 'v' or 'h'")

    sorted_edges = list(sorted(edges, key=itemgetter(min_prop)))
    joined = [ sorted_edges[0] ]
    for e in sorted_edges[1:]:
        last = joined[-1]
        if e[min_prop] <= (last[max_prop] + tolerance):
            if e[max_prop] > last[max_prop]:
                # Extend current edge to new extremity
                joined[-1] = utils.resize_object(last, max_prop, e[max_prop])
            else:
                # Do nothing; edge is fully contained in pre-existing edge
                pass
        else:
            # Edge is separate from previous edges
            joined.append(e)

    return joined

def merge_edges(edges, snap_tolerance, join_tolerance):
    """
    Using the `snap_edges` and `join_edge_group` methods above, merge a list of edges into a more "seamless" list.
    """
    def get_group(edge):
        if edge["orientation"] == "h":
            return ("h", edge["top"])
        else:
            return ("v", edge["x0"])

    if snap_tolerance > 0:
        edges = snap_edges(edges, snap_tolerance)

    if join_tolerance > 0:
        _sorted = sorted(edges, key=get_group)
        edge_groups = itertools.groupby(_sorted, key=get_group)
        edge_gen = (join_edge_group(items, k[0], join_tolerance)
            for k, items in edge_groups)
        edges = list(itertools.chain(*edge_gen))
    return edges

def words_to_edges_h(words,
    word_threshold=DEFAULT_MIN_WORDS_HORIZONTAL):
    """
    Find (imaginary) horizontal lines that connect the tops of at least `word_threshold` words.
    """
    by_top = utils.cluster_objects(words, "top", 1)
    large_clusters = filter(lambda x: len(x) >= word_threshold, by_top)
    rects = list(map(utils.objects_to_rect, large_clusters))
    if len(rects) == 0:
        return []
    min_x0 = min(map(itemgetter("x0"), rects))
    max_x1 = max(map(itemgetter("x1"), rects))
    edges = [ {
        "x0": min_x0,
        "x1": max_x1,
        "top": r["top"],
        "bottom": r["top"],
        "width": max_x1 - min_x0,
        "orientation": "h"
    } for r in rects ] + [ {
        "x0": min_x0,
        "x1": max_x1,
        "top": r["bottom"],
        "bottom": r["bottom"],
        "width": max_x1 - min_x0,
        "orientation": "h"
    } for r in rects ]

    return edges

def words_to_edges_v(words,
    word_threshold=DEFAULT_MIN_WORDS_VERTICAL):
    """
    Find (imaginary) vertical lines that connect the left, right, or center of at least `word_threshold` words.
    """
    # Find words that share the same left, right, or centerpoints
    by_x0 = utils.cluster_objects(words, "x0", 1)
    by_x1 = utils.cluster_objects(words, "x1", 1)
    by_center = utils.cluster_objects(words, lambda x: (x["x0"] + x["x1"])/2, 1)
    clusters = by_x0 + by_x1 + by_center
    
    # Find the points that align with the most words
    sorted_clusters = sorted(clusters, key=lambda x: -len(x))
    large_clusters = filter(lambda x: len(x) >= word_threshold, sorted_clusters)
    
    # For each of those points, find the rectangles fitting all matching words
    rects = list(map(utils.objects_to_rect, large_clusters))
    
    # Iterate through those rectangles, condensing overlapping rectangles
    condensed_rects = []
    for rect in rects:
        overlap = False
        for c in condensed_rects:
            if utils.objects_overlap(rect, c):
                overlap = True
                break
        if overlap == False:
            condensed_rects.append(rect)
            
    if len(condensed_rects) == 0:
        return []
    sorted_rects = list(sorted(condensed_rects, key=itemgetter("x0")))

    # Find the far-right boundary of the rightmost rectangle
    last_rect = sorted_rects[-1]
    while True:
        words_inside = utils.intersects_bbox(
            [ w for w in words if w["x0"] >= last_rect["x0"] ],
            (last_rect["x0"], last_rect["top"], last_rect["x1"], last_rect["bottom"]), 
        )
        rect = utils.objects_to_rect(words_inside)
        if rect == last_rect:
            break
        else:
            last_rect = rect
    
    # Describe all the left-hand edges of each text cluster
    edges = [ {
        "x0": b["x0"],
        "x1": b["x0"],
        "top": b["top"],
        "bottom": b["bottom"],
        "height": b["bottom"] - b["top"],
        "orientation": "v"
    } for b in sorted_rects ] + [ {
        "x0": last_rect["x1"],
        "x1": last_rect["x1"],
        "top": last_rect["top"],
        "bottom": last_rect["bottom"],
        "height": last_rect["bottom"] - last_rect["top"],
        "orientation": "v"
    } ]
    
    return edges

def edges_to_intersections(edges, x_tolerance=1, y_tolerance=1):
    """
    Given a list of edges, return the points at which they intersect within `tolerance` pixels.
    """
    intersections = {}
    v_edges, h_edges = [ list(filter(lambda x: x["orientation"] == o, edges))
        for o in ("v", "h") ]
    for v in sorted(v_edges, key=itemgetter("x0", "top")):
        for h in sorted(h_edges, key=itemgetter("top", "x0")):
            if ((v["top"] <= (h["top"] + y_tolerance)) and
                (v["bottom"] >= (h["top"] - y_tolerance)) and
                (v["x0"] >= (h["x0"] - x_tolerance)) and
                (v["x0"] <= (h["x1"] + x_tolerance))):
                vertex = (v["x0"], h["top"])
                if vertex not in intersections:
                    intersections[vertex] = { "v": [], "h": [] }
                intersections[vertex]["v"].append(v)
                intersections[vertex]["h"].append(h)
    return intersections
    

def intersections_to_cells(intersections):
    """
    Given a list of points (`intersections`), return all retangular "cells" those points describe.

    `intersections` should be a dictionary with (x0, top) tuples as keys,
    and a list of edge objects as values. The edge objects should correspond
    to the edges that touch the intersection.
    """

    def edge_connects(p1, p2):
        def edges_to_set(edges):
            return set(map(utils.obj_to_bbox, edges))

        if p1[0] == p2[0]:
            common = edges_to_set(intersections[p1]["v"])\
                .intersection(edges_to_set(intersections[p2]["v"]))
            if len(common): return True

        if p1[1] == p2[1]:
            common = edges_to_set(intersections[p1]["h"])\
                .intersection(edges_to_set(intersections[p2]["h"]))
            if len(common): return True
        return False
    
    points = list(sorted(intersections.keys()))     
    n_points = len(points)
    def find_smallest_cell(points, i):
        if i == n_points - 1: return None
        pt = points[i]
        rest = points[i+1:]
        # Get all the points directly below and directly right
        below = [ x for x in rest if x[0] == pt[0] ] 
        right = [ x for x in rest if x[1] == pt[1] ] 
        for below_pt in below:
            if not edge_connects(pt, below_pt): continue
                
            for right_pt in right:
                if not edge_connects(pt, right_pt): continue
                
                bottom_right = (right_pt[0], below_pt[1])
                
                if ((bottom_right in intersections) and
                    edge_connects(bottom_right, right_pt) and
                    edge_connects(bottom_right, below_pt)):
                    
                    return (
                        pt[0],
                        pt[1],
                        bottom_right[0],
                        bottom_right[1]
                    )

    cell_gen = (find_smallest_cell(points, i) for i in range(len(points)))
    return list(filter(None, cell_gen))

def cells_to_tables(cells):
    """
    Given a list of bounding boxes (`cells`), return a list of tables that hold those those cells most simply (and contiguously).
    """
    def bbox_to_corners(bbox):
        x0, top, x1, bottom = bbox
        return list(itertools.product((x0, x1), (top, bottom)))
    
    cells = [ {
        "available": True,
        "bbox": bbox,
        "corners": bbox_to_corners(bbox)
    } for bbox in cells ]
    
        # Iterate through the cells found above, and assign them
    # to contiguous tables
    
    def init_new_table():
        return { "corners": set([]), "cells": [] }
    
    def assign_cell(cell, table):
        table["corners"] = table["corners"].union(set(cell["corners"]))
        table["cells"].append(cell["bbox"])
        cell["available"] = False

    n_cells = len(cells)
    n_assigned = 0
    tables = []
    current_table = init_new_table()
    while True:
        initial_cell_count = len(current_table["cells"])
        for i, cell in enumerate(filter(itemgetter("available"), cells)):
            if len(current_table["cells"]) == 0:
                assign_cell(cell, current_table)
                n_assigned += 1
            else:
                corner_count = sum(c in current_table["corners"]
                    for c in cell["corners"])
                if corner_count > 0 and cell["available"]:
                    assign_cell(cell, current_table)
                    n_assigned += 1
        if n_assigned == n_cells:
            break
        if len(current_table["cells"]) == initial_cell_count:
            tables.append(current_table)
            current_table = init_new_table()

    if len(current_table["cells"]):
        tables.append(current_table)
        
    _sorted = sorted(tables, key=lambda t: min(t["corners"]))
    filtered = [ t["cells"] for t in _sorted if len(t["cells"]) > 1 ]
    return filtered

class CellGroup(object):
    def __init__(self, cells):
        self.cells = cells
        self.bbox = (
            min(map(itemgetter(0), filter(None, cells))),
            min(map(itemgetter(1), filter(None, cells))),
            max(map(itemgetter(2), filter(None, cells))),
            max(map(itemgetter(3), filter(None, cells))),
        )

class Row(CellGroup):
    pass

class Table(object):
    def __init__(self, page, cells):
        self.page = page
        self.cells = cells
        self.bbox = (
            min(map(itemgetter(0), cells)),
            min(map(itemgetter(1), cells)),
            max(map(itemgetter(2), cells)),
            max(map(itemgetter(3), cells)),
        )

    @property
    def rows(self):
        _sorted = sorted(self.cells, key=itemgetter(1, 0)) 
        xs = list(sorted(set(map(itemgetter(0), self.cells))))
        rows = []
        for y, row_cells in itertools.groupby(_sorted, itemgetter(1)):
            xdict = dict((cell[0], cell) for cell in row_cells)
            row = Row([ xdict.get(x) for x in xs ])
            rows.append(row)
        return rows

    def extract(self,
        x_tolerance=utils.DEFAULT_X_TOLERANCE,
        y_tolerance=utils.DEFAULT_Y_TOLERANCE):

        chars = self.page.chars
        table_arr = []
        def char_in_bbox(char, bbox):
            v_mid = (char["top"] + char["bottom"]) / 2
            h_mid = (char["x0"] + char["x1"]) / 2
            x0, top, x1, bottom = bbox
            return (
                (h_mid >= x0) and
                (h_mid < x1) and
                (v_mid >= top) and
                (v_mid < bottom)
            )

        for row in self.rows:
            arr = []
            row_chars = [ char for char in chars
                if char_in_bbox(char, row.bbox) ]

            for cell in row.cells:
                if cell == None:
                    cell_text = None
                else:
                    cell_chars = [ char for char in row_chars
                        if char_in_bbox(char, cell) ]

                    if len(cell_chars):
                        cell_text = utils.extract_text(cell_chars,
                            x_tolerance=x_tolerance,
                            y_tolerance=y_tolerance).strip()
                    else:
                        cell_text = ""
                arr.append(cell_text)
            table_arr.append(arr)

        return table_arr

TABLE_STRATEGIES = [ "lines", "lines_strict", "text", "explicit" ]
DEFAULT_TABLE_SETTINGS = {
    "vertical_strategy": "lines",
    "horizontal_strategy": "lines",
    "explicit_vertical_lines": [],
    "explicit_horizontal_lines": [],
    "snap_tolerance": DEFAULT_SNAP_TOLERANCE,
    "join_tolerance": DEFAULT_JOIN_TOLERANCE,
    "edge_min_length": 3,
    "min_words_vertical": DEFAULT_MIN_WORDS_VERTICAL,
    "min_words_horizontal": DEFAULT_MIN_WORDS_HORIZONTAL,
    "keep_blank_chars": False,
    "text_tolerance": 3,
    "text_x_tolerance": None,
    "text_y_tolerance": None,
    "intersection_tolerance": 3,
    "intersection_x_tolerance": None,
    "intersection_y_tolerance": None,
}

class TableFinder(object):
    """
    Given a PDF page, find plausible table structures.

    Largely borrowed from Anssi Nurminen's master's thesis: http://dspace.cc.tut.fi/dpub/bitstream/handle/123456789/21520/Nurminen.pdf?sequence=3

    ... and inspired by Tabula: https://github.com/tabulapdf/tabula-extractor/issues/16
    """
    def __init__(self, page, settings={}):
        for k in settings.keys():
            if k not in DEFAULT_TABLE_SETTINGS:
                raise ValueError("Unrecognized table setting: '{0}'".format(
                    k
                ))
        self.page = page
        self.settings = dict(DEFAULT_TABLE_SETTINGS)
        self.settings.update(settings)
        for var, fallback in [
            ("text_x_tolerance", "text_tolerance"),
            ("text_y_tolerance", "text_tolerance"),
            ("intersection_x_tolerance", "intersection_tolerance"),
            ("intersection_y_tolerance", "intersection_tolerance"),
        ]:
            if self.settings[var] == None:
                self.settings.update({
                    var: self.settings[fallback]
                })
        self.edges = self.get_edges()
        self.intersections = edges_to_intersections(
            self.edges,
            self.settings["intersection_x_tolerance"],
            self.settings["intersection_y_tolerance"],
        )
        self.cells = intersections_to_cells(
            self.intersections
        )
        self.tables = [ Table(self.page, t)
            for t in cells_to_tables(self.cells) ]

    def get_edges(self):
        settings = self.settings
        for name in [ "vertical", "horizontal" ]:
            strategy = settings[name + "_strategy"]
            if strategy not in TABLE_STRATEGIES:
                raise ValueError("{0} must be one of {{{1}}}".format(
                    name + "_strategy",
                    ",".join(TABLE_STRATEGIES)
                ))
            if strategy == "explicit":
                if len(settings["explicit_" + name + "_lines"]) < 2:
                    raise ValueError("If {0} == 'explicit', {1} must be specified as list/tuple of two or more floats/ints.".format(
                        strategy + "_strategy",
                        "explicit_" + name + "_lines",
                    ))

        v_strat = settings["vertical_strategy"]
        h_strat = settings["horizontal_strategy"]

        if v_strat == "text" or h_strat == "text":
            xt = settings["text_x_tolerance"]
            if xt == None:
                xt = settings["text_tolerance"]
            yt = settings["text_y_tolerance"]
            if yt == None:
                yt = settings["text_tolerance"]
            words = self.page.extract_words(
                x_tolerance=xt,
                y_tolerance=yt,
                keep_blank_chars=settings["keep_blank_chars"]
            )

        def v_edge_desc_to_edge(desc):
            if isinstance(desc, dict):
                edge = {
                    "x0": desc.get("x0", desc.get("x")),
                    "x1": desc.get("x1", desc.get("x")),
                    "top": desc.get("top", self.page.bbox[1]),
                    "bottom": desc.get("bottom", self.page.bbox[3]),
                    "orientation": "v"
                }
            else:
                edge = {
                    "x0": desc,
                    "x1": desc,
                    "top": self.page.bbox[1],
                    "bottom": self.page.bbox[3],
                }
            edge["height"] = edge["bottom"] - edge["top"]
            edge["orientation"] = "v"
            return edge

        v_explicit = list(map(v_edge_desc_to_edge, settings["explicit_vertical_lines"]))

        if  v_strat == "lines":
            v_base = utils.filter_edges(self.page.edges, "v")
        elif v_strat == "lines_strict":
            v_base = utils.filter_edges(self.page.edges, "v",
                edge_type="lines")
        elif v_strat == "text":
            v_base = words_to_edges_v(words,
                word_threshold=settings["min_words_vertical"])
        elif v_strat == "explicit":
            v_base = []

        v = v_base + v_explicit
        
        def h_edge_desc_to_edge(desc):
            if isinstance(desc, dict):
                edge = {
                    "x0": desc.get("x0", self.page.bbox[0]),
                    "x1": desc.get("x1", self.page.bbox[2]),
                    "top": desc.get("top", desc.get("bottom")),
                    "bottom": desc.get("bottom", desc.get("top")),
                }
            else:
                edge = {
                    "x0": self.page.bbox[0],
                    "x1": self.page.bbox[2],
                    "top": desc,
                    "bottom": desc,
                }
            edge["width"] = edge["x1"] - edge["x0"]
            edge["orientation"] = "h"
            return edge

        h_explicit = list(map(h_edge_desc_to_edge, settings["explicit_horizontal_lines"]))

        if  h_strat == "lines":
            h_base = utils.filter_edges(self.page.edges, "h")
        elif h_strat == "lines_strict":
            h_base = utils.filter_edges(self.page.edges, "h",
                edge_type="lines")
        elif h_strat == "text":
            h_base = words_to_edges_h(words,
                word_threshold=settings["min_words_horizontal"])
        elif h_strat == "explicit":
            h_base = []

        h = h_base + h_explicit

        edges = list(v) + list(h)
        if settings["snap_tolerance"] > 0 or settings["join_tolerance"] > 0:
            edges = merge_edges(edges,
                snap_tolerance=settings["snap_tolerance"],
                join_tolerance=settings["join_tolerance"],
            )
        return utils.filter_edges(edges,
            min_length=settings["edge_min_length"])



import pdfminer

#def list_value(x):
#    x = pdfminer.pdftypes.resolve1(x)
#    if not isinstance(x, (list, tuple)):
#        print(x, type(x))
#        return []
#    return x
#
#pdfminer.pdffont.list_value = list_value

#pdfminer.settings.STRICT = False
#pdfminer.pdftypes.STRICT = False
pdfminer.pdfinterp.STRICT = False



from pdfminer.pdffont import PDFFont
from pdfminer.pdftypes import stream_value

def get_fontfile(self):
    desc = self.descriptor
    ff_key = None
    for key in [ "FontFile", "FontFile2", "FontFile3" ]:
        if key in desc:
            ff_key = key
            break
    if ff_key is not None:
        return stream_value(desc[ff_key]).get_data()
    
PDFFont.get_fontfile = get_fontfile
#import sys
#import struct
#from io import BytesIO
#from pdfminer.cmapdb import CMapDB
#from pdfminer.cmapdb import CMapParser
#from pdfminer.cmapdb import FileUnicodeMap
#from pdfminer.cmapdb import CMap
#from pdfminer.encodingdb import EncodingDB
#from pdfminer.encodingdb import name2unicode
#from pdfminer.psparser import PSStackParser
#from pdfminer.psparser import PSEOF
#from pdfminer.psparser import LIT
#from pdfminer.psparser import KWD
#from pdfminer import settings
#from pdfminer.psparser import PSLiteral
#from pdfminer.psparser import literal_name
#from pdfminer.pdftypes import PDFException
#from pdfminer.pdftypes import resolve1
#from pdfminer.pdftypes import int_value
#from pdfminer.pdftypes import num_value
#from pdfminer.pdftypes import list_value
#from pdfminer.pdftypes import dict_value
#from pdfminer.pdftypes import stream_value
#from pdfminer.fontmetrics import FONT_METRICS
#from pdfminer.utils import apply_matrix_norm
#from pdfminer.utils import nunpack
#from pdfminer.utils import choplist
#from pdfminer.utils import isnumber
#
#import six #Python 2+3 compatibility
#
#
#def get_widths(seq):
#    widths = {}
#    r = []
#    for v in seq:
#        if isinstance(v, list):
#            if r:
#                char1 = r[-1]
#                for (i, w) in enumerate(v):
#                    widths[char1+i] = w
#                r = []
#        elif isnumber(v):
#            r.append(v)
#            if len(r) == 3:
#                (char1, char2, w) = r
#                for i in range(char1, char2+1):
#                    widths[i] = w
#                r = []
#    return widths
##assert get_widths([1]) == {}
##assert get_widths([1,2,3]) == {1:3, 2:3}
##assert get_widths([1,[2,3],6,[7,8]]) == {1:2,2:3, 6:7,7:8}
#
#
#def get_widths2(seq):
#    widths = {}
#    r = []
#    for v in seq:
#        if isinstance(v, list):
#            if r:
#                char1 = r[-1]
#                for (i, (w, vx, vy)) in enumerate(choplist(3, v)):
#                    widths[char1+i] = (w, (vx, vy))
#                r = []
#        elif isnumber(v):
#            r.append(v)
#            if len(r) == 5:
#                (char1, char2, w, vx, vy) = r
#                for i in range(char1, char2+1):
#                    widths[i] = (w, (vx, vy))
#                r = []
#    return widths
##assert get_widths2([1]) == {}
##assert get_widths2([1,2,3,4,5]) == {1:(3, (4,5)), 2:(3, (4,5))}
##assert get_widths2([1,[2,3,4,5],6,[7,8,9]]) == {1:(2, (3,4)), 6:(7, (8,9))}
#
#
###  FontMetricsDB
###
#class FontMetricsDB(object):
#
#    @classmethod
#    def get_metrics(klass, fontname):
#        return FONT_METRICS[fontname]
#
#
###  Type1FontHeaderParser
###
#class Type1FontHeaderParser(PSStackParser):
#
#    KEYWORD_BEGIN = KWD(b'begin')
#    KEYWORD_END = KWD(b'end')
#    KEYWORD_DEF = KWD(b'def')
#    KEYWORD_PUT = KWD(b'put')
#    KEYWORD_DICT = KWD(b'dict')
#    KEYWORD_ARRAY = KWD(b'array')
#    KEYWORD_READONLY = KWD(b'readonly')
#    KEYWORD_FOR = KWD(b'for')
#    KEYWORD_FOR = KWD(b'for')
#
#    def __init__(self, data):
#        PSStackParser.__init__(self, data)
#        self._cid2unicode = {}
#        return
#
#    def get_encoding(self):
#        while 1:
#            try:
#                (cid, name) = self.nextobject()
#            except PSEOF:
#                break
#            try:
#                self._cid2unicode[cid] = name2unicode(name)
#            except KeyError:
#                pass
#        return self._cid2unicode
#
#    def do_keyword(self, pos, token):
#        if token is self.KEYWORD_PUT:
#            ((_, key), (_, value)) = self.pop(2)
#            if (isinstance(key, int) and
#                isinstance(value, PSLiteral)):
#                self.add_results((key, literal_name(value)))
#        return
#
#
#NIBBLES = ('0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '.', 'e', 'e-', None, '-')
#
#
###  CFFFont
###  (Format specified in Adobe Technical Note: #5176
###   "The Compact Font Format Specification")
###
#def getdict(data):
#    d = {}
#    fp = BytesIO(data)
#    stack = []
#    while 1:
#        c = fp.read(1)
#        if not c:
#            break
#        b0 = ord(c)
#        if b0 <= 21:
#            d[b0] = stack
#            stack = []
#            continue
#        if b0 == 30:
#            s = ''
#            loop = True
#            while loop:
#                b = ord(fp.read(1))
#                for n in (b >> 4, b & 15):
#                    if n == 15:
#                        loop = False
#                    else:
#                        s += NIBBLES[n]
#            value = float(s)
#        elif 32 <= b0 and b0 <= 246:
#            value = b0-139
#        else:
#            b1 = ord(fp.read(1))
#            if 247 <= b0 and b0 <= 250:
#                value = ((b0-247) << 8)+b1+108
#            elif 251 <= b0 and b0 <= 254:
#                value = -((b0-251) << 8)-b1-108
#            else:
#                b2 = ord(fp.read(1))
#                if 128 <= b1:
#                    b1 -= 256
#                if b0 == 28:
#                    value = b1 << 8 | b2
#                else:
#                    value = b1 << 24 | b2 << 16 | struct.unpack('>H', fp.read(2))[0]
#        stack.append(value)
#    return d
#
#
#class CFFFont(object):
#
#    STANDARD_STRINGS = (
#      '.notdef', 'space', 'exclam', 'quotedbl', 'numbersign',
#      'dollar', 'percent', 'ampersand', 'quoteright', 'parenleft',
#      'parenright', 'asterisk', 'plus', 'comma', 'hyphen', 'period',
#      'slash', 'zero', 'one', 'two', 'three', 'four', 'five', 'six',
#      'seven', 'eight', 'nine', 'colon', 'semicolon', 'less', 'equal',
#      'greater', 'question', 'at', 'A', 'B', 'C', 'D', 'E', 'F', 'G',
#      'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
#      'U', 'V', 'W', 'X', 'Y', 'Z', 'bracketleft', 'backslash',
#      'bracketright', 'asciicircum', 'underscore', 'quoteleft', 'a',
#      'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
#      'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
#      'braceleft', 'bar', 'braceright', 'asciitilde', 'exclamdown',
#      'cent', 'sterling', 'fraction', 'yen', 'florin', 'section',
#      'currency', 'quotesingle', 'quotedblleft', 'guillemotleft',
#      'guilsinglleft', 'guilsinglright', 'fi', 'fl', 'endash',
#      'dagger', 'daggerdbl', 'periodcentered', 'paragraph', 'bullet',
#      'quotesinglbase', 'quotedblbase', 'quotedblright',
#      'guillemotright', 'ellipsis', 'perthousand', 'questiondown',
#      'grave', 'acute', 'circumflex', 'tilde', 'macron', 'breve',
#      'dotaccent', 'dieresis', 'ring', 'cedilla', 'hungarumlaut',
#      'ogonek', 'caron', 'emdash', 'AE', 'ordfeminine', 'Lslash',
#      'Oslash', 'OE', 'ordmasculine', 'ae', 'dotlessi', 'lslash',
#      'oslash', 'oe', 'germandbls', 'onesuperior', 'logicalnot', 'mu',
#      'trademark', 'Eth', 'onehalf', 'plusminus', 'Thorn',
#      'onequarter', 'divide', 'brokenbar', 'degree', 'thorn',
#      'threequarters', 'twosuperior', 'registered', 'minus', 'eth',
#      'multiply', 'threesuperior', 'copyright', 'Aacute',
#      'Acircumflex', 'Adieresis', 'Agrave', 'Aring', 'Atilde',
#      'Ccedilla', 'Eacute', 'Ecircumflex', 'Edieresis', 'Egrave',
#      'Iacute', 'Icircumflex', 'Idieresis', 'Igrave', 'Ntilde',
#      'Oacute', 'Ocircumflex', 'Odieresis', 'Ograve', 'Otilde',
#      'Scaron', 'Uacute', 'Ucircumflex', 'Udieresis', 'Ugrave',
#      'Yacute', 'Ydieresis', 'Zcaron', 'aacute', 'acircumflex',
#      'adieresis', 'agrave', 'aring', 'atilde', 'ccedilla', 'eacute',
#      'ecircumflex', 'edieresis', 'egrave', 'iacute', 'icircumflex',
#      'idieresis', 'igrave', 'ntilde', 'oacute', 'ocircumflex',
#      'odieresis', 'ograve', 'otilde', 'scaron', 'uacute',
#      'ucircumflex', 'udieresis', 'ugrave', 'yacute', 'ydieresis',
#      'zcaron', 'exclamsmall', 'Hungarumlautsmall', 'dollaroldstyle',
#      'dollarsuperior', 'ampersandsmall', 'Acutesmall',
#      'parenleftsuperior', 'parenrightsuperior', 'twodotenleader',
#      'onedotenleader', 'zerooldstyle', 'oneoldstyle', 'twooldstyle',
#      'threeoldstyle', 'fouroldstyle', 'fiveoldstyle', 'sixoldstyle',
#      'sevenoldstyle', 'eightoldstyle', 'nineoldstyle',
#      'commasuperior', 'threequartersemdash', 'periodsuperior',
#      'questionsmall', 'asuperior', 'bsuperior', 'centsuperior',
#      'dsuperior', 'esuperior', 'isuperior', 'lsuperior', 'msuperior',
#      'nsuperior', 'osuperior', 'rsuperior', 'ssuperior', 'tsuperior',
#      'ff', 'ffi', 'ffl', 'parenleftinferior', 'parenrightinferior',
#      'Circumflexsmall', 'hyphensuperior', 'Gravesmall', 'Asmall',
#      'Bsmall', 'Csmall', 'Dsmall', 'Esmall', 'Fsmall', 'Gsmall',
#      'Hsmall', 'Ismall', 'Jsmall', 'Ksmall', 'Lsmall', 'Msmall',
#      'Nsmall', 'Osmall', 'Psmall', 'Qsmall', 'Rsmall', 'Ssmall',
#      'Tsmall', 'Usmall', 'Vsmall', 'Wsmall', 'Xsmall', 'Ysmall',
#      'Zsmall', 'colonmonetary', 'onefitted', 'rupiah', 'Tildesmall',
#      'exclamdownsmall', 'centoldstyle', 'Lslashsmall', 'Scaronsmall',
#      'Zcaronsmall', 'Dieresissmall', 'Brevesmall', 'Caronsmall',
#      'Dotaccentsmall', 'Macronsmall', 'figuredash', 'hypheninferior',
#      'Ogoneksmall', 'Ringsmall', 'Cedillasmall', 'questiondownsmall',
#      'oneeighth', 'threeeighths', 'fiveeighths', 'seveneighths',
#      'onethird', 'twothirds', 'zerosuperior', 'foursuperior',
#      'fivesuperior', 'sixsuperior', 'sevensuperior', 'eightsuperior',
#      'ninesuperior', 'zeroinferior', 'oneinferior', 'twoinferior',
#      'threeinferior', 'fourinferior', 'fiveinferior', 'sixinferior',
#      'seveninferior', 'eightinferior', 'nineinferior',
#      'centinferior', 'dollarinferior', 'periodinferior',
#      'commainferior', 'Agravesmall', 'Aacutesmall',
#      'Acircumflexsmall', 'Atildesmall', 'Adieresissmall',
#      'Aringsmall', 'AEsmall', 'Ccedillasmall', 'Egravesmall',
#      'Eacutesmall', 'Ecircumflexsmall', 'Edieresissmall',
#      'Igravesmall', 'Iacutesmall', 'Icircumflexsmall',
#      'Idieresissmall', 'Ethsmall', 'Ntildesmall', 'Ogravesmall',
#      'Oacutesmall', 'Ocircumflexsmall', 'Otildesmall',
#      'Odieresissmall', 'OEsmall', 'Oslashsmall', 'Ugravesmall',
#      'Uacutesmall', 'Ucircumflexsmall', 'Udieresissmall',
#      'Yacutesmall', 'Thornsmall', 'Ydieresissmall', '001.000',
#      '001.001', '001.002', '001.003', 'Black', 'Bold', 'Book',
#      'Light', 'Medium', 'Regular', 'Roman', 'Semibold',
#    )
#
#    class INDEX(object):
#
#        def __init__(self, fp):
#            self.fp = fp
#            self.offsets = []
#            (count, offsize) = struct.unpack('>HB', self.fp.read(3))
#            for i in range(count+1):
#                self.offsets.append(nunpack(self.fp.read(offsize)))
#            self.base = self.fp.tell()-1
#            self.fp.seek(self.base+self.offsets[-1])
#            return
#
#        def __repr__(self):
#            return '<INDEX: size=%d>' % len(self)
#
#        def __len__(self):
#            return len(self.offsets)-1
#
#        def __getitem__(self, i):
#            self.fp.seek(self.base+self.offsets[i])
#            return self.fp.read(self.offsets[i+1]-self.offsets[i])
#
#        def __iter__(self):
#            return iter(self[i] for i in range(len(self)))
#
#    def __init__(self, name, fp):
#        self.name = name
#        self.fp = fp
#        # Header
#        (_major, _minor, hdrsize, offsize) = struct.unpack('BBBB', self.fp.read(4))
#        self.fp.read(hdrsize-4)
#        # Name INDEX
#        self.name_index = self.INDEX(self.fp)
#        # Top DICT INDEX
#        self.dict_index = self.INDEX(self.fp)
#        # String INDEX
#        self.string_index = self.INDEX(self.fp)
#        # Global Subr INDEX
#        self.subr_index = self.INDEX(self.fp)
#        # Top DICT DATA
#        self.top_dict = getdict(self.dict_index[0])
#        (charset_pos,) = self.top_dict.get(15, [0])
#        (encoding_pos,) = self.top_dict.get(16, [0])
#        (charstring_pos,) = self.top_dict.get(17, [0])
#        # CharStrings
#        self.fp.seek(charstring_pos)
#        self.charstring = self.INDEX(self.fp)
#        self.nglyphs = len(self.charstring)
#        # Encodings
#        self.code2gid = {}
#        self.gid2code = {}
#        self.fp.seek(encoding_pos)
#        format = self.fp.read(1)
#        if format == b'\x00':
#            # Format 0
#            (n,) = struct.unpack('B', self.fp.read(1))
#            for (code, gid) in enumerate(struct.unpack('B'*n, self.fp.read(n))):
#                self.code2gid[code] = gid
#                self.gid2code[gid] = code
#        elif format == b'\x01':
#            # Format 1
#            (n,) = struct.unpack('B', self.fp.read(1))
#            code = 0
#            for i in range(n):
#                (first, nleft) = struct.unpack('BB', self.fp.read(2))
#                for gid in range(first, first+nleft+1):
#                    self.code2gid[code] = gid
#                    self.gid2code[gid] = code
#                    code += 1
#        else:
#            raise ValueError('unsupported encoding format: %r' % format)
#        # Charsets
#        self.name2gid = {}
#        self.gid2name = {}
#        self.fp.seek(charset_pos)
#        format = self.fp.read(1)
#        if format == b'\x00':
#            # Format 0
#            n = self.nglyphs-1
#            for (gid, sid) in enumerate(struct.unpack('>'+'H'*n, self.fp.read(2*n))):
#                gid += 1
#                name = self.getstr(sid)
#                self.name2gid[name] = gid
#                self.gid2name[gid] = name
#        elif format == b'\x01':
#            # Format 1
#            (n,) = struct.unpack('B', self.fp.read(1))
#            sid = 0
#            for i in range(n):
#                (first, nleft) = struct.unpack('BB', self.fp.read(2))
#                for gid in range(first, first+nleft+1):
#                    name = self.getstr(sid)
#                    self.name2gid[name] = gid
#                    self.gid2name[gid] = name
#                    sid += 1
#        elif format == b'\x02':
#            # Format 2
#            assert 0
#        else:
#            raise ValueError('unsupported charset format: %r' % format)
#        #print self.code2gid
#        #print self.name2gid
#        #assert 0
#        return
#
#    def getstr(self, sid):
#        if sid < len(self.STANDARD_STRINGS):
#            return self.STANDARD_STRINGS[sid]
#        return self.string_index[sid-len(self.STANDARD_STRINGS)]
#
#
###  TrueTypeFont
###
#class TrueTypeFont(object):
#
#    class CMapNotFound(Exception):
#        pass
#
#    def __init__(self, name, fp):
#        self.name = name
#        self.fp = fp
#        self.tables = {}
#        self.fonttype = fp.read(4)
#        (ntables, _1, _2, _3) = struct.unpack('>HHHH', fp.read(8))
#        for _ in range(ntables):
#            (name, tsum, offset, length) = struct.unpack('>4sLLL', fp.read(16))
#            self.tables[name] = (offset, length)
#        return
#
#    def create_unicode_map(self):
#        if 'cmap' not in self.tables:
#            raise TrueTypeFont.CMapNotFound
#        (base_offset, length) = self.tables['cmap']
#        fp = self.fp
#        fp.seek(base_offset)
#        (version, nsubtables) = struct.unpack('>HH', fp.read(4))
#        subtables = []
#        for i in range(nsubtables):
#            subtables.append(struct.unpack('>HHL', fp.read(8)))
#        char2gid = {}
#        # Only supports subtable type 0, 2 and 4.
#        for (_1, _2, st_offset) in subtables:
#            fp.seek(base_offset+st_offset)
#            (fmttype, fmtlen, fmtlang) = struct.unpack('>HHH', fp.read(6))
#            if fmttype == 0:
#                char2gid.update(enumerate(struct.unpack('>256B', fp.read(256))))
#            elif fmttype == 2:
#                subheaderkeys = struct.unpack('>256H', fp.read(512))
#                firstbytes = [0]*8192
#                for (i, k) in enumerate(subheaderkeys):
#                    firstbytes[k//8] = i
#                nhdrs = max(subheaderkeys)//8 + 1
#                hdrs = []
#                for i in range(nhdrs):
#                    (firstcode, entcount, delta, offset) = struct.unpack('>HHhH', fp.read(8))
#                    hdrs.append((i, firstcode, entcount, delta, fp.tell()-2+offset))
#                for (i, firstcode, entcount, delta, pos) in hdrs:
#                    if not entcount:
#                        continue
#                    first = firstcode + (firstbytes[i] << 8)
#                    fp.seek(pos)
#                    for c in range(entcount):
#                        gid = struct.unpack('>H', fp.read(2))
#                        if gid:
#                            gid += delta
#                        char2gid[first+c] = gid
#            elif fmttype == 4:
#                (segcount, _1, _2, _3) = struct.unpack('>HHHH', fp.read(8))
#                segcount //= 2
#                ecs = struct.unpack('>%dH' % segcount, fp.read(2*segcount))
#                fp.read(2)
#                scs = struct.unpack('>%dH' % segcount, fp.read(2*segcount))
#                idds = struct.unpack('>%dh' % segcount, fp.read(2*segcount))
#                pos = fp.tell()
#                idrs = struct.unpack('>%dH' % segcount, fp.read(2*segcount))
#                for (ec, sc, idd, idr) in zip(ecs, scs, idds, idrs):
#                    if idr:
#                        fp.seek(pos+idr)
#                        for c in range(sc, ec+1):
#                            char2gid[c] = (struct.unpack('>H', fp.read(2))[0] + idd) & 0xffff
#                    else:
#                        for c in range(sc, ec+1):
#                            char2gid[c] = (c + idd) & 0xffff
#            else:
#                assert 0
#        # create unicode map
#        unicode_map = FileUnicodeMap()
#        for (char, gid) in char2gid.iteritems():
#            unicode_map.add_cid2unichr(gid, char)
#        return unicode_map
#
#
###  Fonts
###
#class PDFFontError(PDFException):
#    pass
#
#
#class PDFUnicodeNotDefined(PDFFontError):
#    pass
#
#LITERAL_STANDARD_ENCODING = LIT('StandardEncoding')
#LITERAL_TYPE1C = LIT('Type1C')
#
#
## PDFFont
#class PDFFont(object):
#
#    def __init__(self, descriptor, widths, default_width=None):
#        self.descriptor = descriptor
#        self.widths = widths
#        self.fontname = resolve1(descriptor.get('FontName', 'unknown'))
#        if isinstance(self.fontname, PSLiteral):
#            self.fontname = literal_name(self.fontname)
#        self.flags = int_value(descriptor.get('Flags', 0))
#        self.ascent = num_value(descriptor.get('Ascent', 0))
#        self.descent = num_value(descriptor.get('Descent', 0))
#        self.italic_angle = num_value(descriptor.get('ItalicAngle', 0))
#        self.default_width = default_width or num_value(descriptor.get('MissingWidth', 0))
#        self.leading = num_value(descriptor.get('Leading', 0))
#        self.bbox = list_value(descriptor.get('FontBBox', (0, 0, 0, 0)))
#        self.hscale = self.vscale = .001
#        return
#
#    def __repr__(self):
#        return '<PDFFont>'
#
#    def is_vertical(self):
#        return False
#
#    def is_multibyte(self):
#        return False
#
#    def decode(self, bytes):
#        return [six.indexbytes(bytes, i) for i,_ in enumerate(bytes)] # map(ord, bytes)
#
#    def get_ascent(self):
#        return self.ascent * self.vscale
#
#    def get_descent(self):
#        return self.descent * self.vscale
#
#    def get_width(self):
#        w = self.bbox[2]-self.bbox[0]
#        if w == 0:
#            w = -self.default_width
#        return w * self.hscale
#
#    def get_height(self):
#        h = self.bbox[3]-self.bbox[1]
#        if h == 0:
#            h = self.ascent - self.descent
#        return h * self.vscale
#
#    def char_width(self, cid):
#        try:
#            return self.widths[cid] * self.hscale
#        except KeyError:
#            try:
#                return self.widths[self.to_unichr(cid)] * self.hscale
#            except (KeyError, PDFUnicodeNotDefined):
#                return self.default_width * self.hscale
#
#    def char_disp(self, cid):
#        return 0
#
#    def string_width(self, s):
#        return sum(self.char_width(cid) for cid in self.decode(s))
#
#
## PDFSimpleFont
#class PDFSimpleFont(PDFFont):
#
#    def __init__(self, descriptor, widths, spec):
#        # Font encoding is specified either by a name of
#        # built-in encoding or a dictionary that describes
#        # the differences.
#        if 'Encoding' in spec:
#            encoding = resolve1(spec['Encoding'])
#        else:
#            encoding = LITERAL_STANDARD_ENCODING
#        if isinstance(encoding, dict):
#            name = literal_name(encoding.get('BaseEncoding', LITERAL_STANDARD_ENCODING))
#            diff = list_value(encoding.get('Differences', []))
#            self.cid2unicode = EncodingDB.get_encoding(name, diff)
#        else:
#            self.cid2unicode = EncodingDB.get_encoding(literal_name(encoding))
#        self.unicode_map = None
#        if 'ToUnicode' in spec:
#            strm = stream_value(spec['ToUnicode'])
#            self.unicode_map = FileUnicodeMap()
#            CMapParser(self.unicode_map, BytesIO(strm.get_data())).run()
#        PDFFont.__init__(self, descriptor, widths)
#        return
#
#    def to_unichr(self, cid):
#        if self.unicode_map:
#            try:
#                return self.unicode_map.get_unichr(cid)
#            except KeyError:
#                pass
#        try:
#            return self.cid2unicode[cid]
#        except KeyError:
#            raise PDFUnicodeNotDefined(None, cid)
#
#
## PDFType1Font
#class PDFType1Font(PDFSimpleFont):
#
#    def __init__(self, rsrcmgr, spec):
#        try:
#            self.basefont = literal_name(spec['BaseFont'])
#        except KeyError:
#            if settings.STRICT:
#                raise PDFFontError('BaseFont is missing')
#            self.basefont = 'unknown'
#        try:
#            (descriptor, widths) = FontMetricsDB.get_metrics(self.basefont)
#        except KeyError:
#            descriptor = dict_value(spec.get('FontDescriptor', {}))
#            firstchar = int_value(spec.get('FirstChar', 0))
#            #lastchar = int_value(spec.get('LastChar', 255))
#            widths = list_value(spec.get('Widths', [0]*256))
#            widths = dict((i+firstchar, w) for (i, w) in enumerate(widths))
#        PDFSimpleFont.__init__(self, descriptor, widths, spec)
#        if 'Encoding' not in spec and 'FontFile' in descriptor:
#            # try to recover the missing encoding info from the font file.
#            self.fontfile = stream_value(descriptor.get('FontFile'))
#            length1 = int_value(self.fontfile['Length1'])
#            data = self.fontfile.get_data()[:length1]
#            parser = Type1FontHeaderParser(BytesIO(data))
#            self.cid2unicode = parser.get_encoding()
#        return
#
#    def __repr__(self):
#        return '<PDFType1Font: basefont=%r>' % self.basefont
#
#
## PDFTrueTypeFont
#class PDFTrueTypeFont(PDFType1Font):
#
#    def __repr__(self):
#        return '<PDFTrueTypeFont: basefont=%r>' % self.basefont
#
#
## PDFType3Font
#class PDFType3Font(PDFSimpleFont):
#
#    def __init__(self, rsrcmgr, spec):
#        firstchar = int_value(spec.get('FirstChar', 0))
#        #lastchar = int_value(spec.get('LastChar', 0))
#        widths = list_value(spec.get('Widths', [0]*256))
#        widths = dict((i+firstchar, w) for (i, w) in enumerate(widths))
#        if 'FontDescriptor' in spec:
#            descriptor = dict_value(spec['FontDescriptor'])
#        else:
#            descriptor = {'Ascent': 0, 'Descent': 0,
#                          'FontBBox': spec['FontBBox']}
#        PDFSimpleFont.__init__(self, descriptor, widths, spec)
#        self.matrix = tuple(list_value(spec.get('FontMatrix')))
#        (_, self.descent, _, self.ascent) = self.bbox
#        (self.hscale, self.vscale) = apply_matrix_norm(self.matrix, (1, 1))
#        return
#
#    def __repr__(self):
#        return '<PDFType3Font>'
#
#
## PDFCIDFont
#class PDFCIDFont(PDFFont):
#
#    def __init__(self, rsrcmgr, spec, strict=settings.STRICT):
#        try:
#            self.basefont = literal_name(spec['BaseFont'])
#        except KeyError:
#            if strict:
#                raise PDFFontError('BaseFont is missing')
#            self.basefont = 'unknown'
#        self.cidsysteminfo = dict_value(spec.get('CIDSystemInfo', {}))
#        self.cidcoding = '%s-%s' % (resolve1(self.cidsysteminfo.get('Registry', b'unknown')).decode("latin1"),
#                                    resolve1(self.cidsysteminfo.get('Ordering', b'unknown')).decode("latin1"))
#        try:
#            name = literal_name(spec['Encoding'])
#        except KeyError:
#            if strict:
#                raise PDFFontError('Encoding is unspecified')
#            name = 'unknown'
#        try:
#            self.cmap = CMapDB.get_cmap(name)
#        except CMapDB.CMapNotFound as e:
#            if strict:
#                raise PDFFontError(e)
#            self.cmap = CMap()
#        try:
#            descriptor = dict_value(spec['FontDescriptor'])
#        except KeyError:
#            if strict:
#                raise PDFFontError('FontDescriptor is missing')
#            descriptor = {}
#        ttf = None
#        if 'FontFile2' in descriptor:
#            self.fontfile = stream_value(descriptor.get('FontFile2'))
#            ttf = TrueTypeFont(self.basefont,
#                               BytesIO(self.fontfile.get_data()))
#        if 'FontFile3' in descriptor:
#            print(descriptor)
#            self.fontfile = stream_value(descriptor.get('FontFile3'))
#            ttf = TrueTypeFont(self.basefont,
#                               BytesIO(self.fontfile.get_data()))
#        self.unicode_map = None
#        if 'ToUnicode' in spec:
#            strm = stream_value(spec['ToUnicode'])
#            self.unicode_map = FileUnicodeMap()
#            CMapParser(self.unicode_map, BytesIO(strm.get_data())).run()
#        elif self.cidcoding in ('Adobe-Identity', 'Adobe-UCS'):
#            if ttf:
#                try:
#                    self.unicode_map = ttf.create_unicode_map()
#                except TrueTypeFont.CMapNotFound:
#                    pass
#        else:
#            try:
#                self.unicode_map = CMapDB.get_unicode_map(self.cidcoding, self.cmap.is_vertical())
#            except CMapDB.CMapNotFound as e:
#                pass
#
#        self.vertical = self.cmap.is_vertical()
#        if self.vertical:
#            # writing mode: vertical
#            widths = get_widths2(list_value(spec.get('W2', [])))
#            self.disps = dict((cid, (vx, vy)) for (cid, (_, (vx, vy))) in six.iteritems(widths))
#            (vy, w) = spec.get('DW2', [880, -1000])
#            self.default_disp = (None, vy)
#            widths = dict((cid, w) for (cid, (w, _)) in six.iteritems(widths))
#            default_width = w
#        else:
#            # writing mode: horizontal
#            self.disps = {}
#            self.default_disp = 0
#            widths = get_widths(list_value(spec.get('W', [])))
#            default_width = spec.get('DW', 1000)
#        PDFFont.__init__(self, descriptor, widths, default_width=default_width)
#        return
#
#    def __repr__(self):
#        return '<PDFCIDFont: basefont=%r, cidcoding=%r>' % (self.basefont, self.cidcoding)
#
#    def is_vertical(self):
#        return self.vertical
#
#    def is_multibyte(self):
#        return True
#
#    def decode(self, bytes):
#        return self.cmap.decode(bytes)
#
#    def char_disp(self, cid):
#        "Returns an integer for horizontal fonts, a tuple for vertical fonts."
#        return self.disps.get(cid, self.default_disp)
#
#    def to_unichr(self, cid):
#        try:
#            if not self.unicode_map:
#                raise KeyError(cid)
#            return self.unicode_map.get_unichr(cid)
#        except KeyError:
#            raise PDFUnicodeNotDefined(self.cidcoding, cid)
#
#import pdfminer
#pdfminer.pdffont.CFFFont = CFFFont
#pdfminer.pdffont.TrueTypeFont = TrueTypeFont
#pdfminer.pdffont.PDFFontError = PDFFontError
#pdfminer.pdffont.PDFUnicodeNotDefined = PDFUnicodeNotDefined
#pdfminer.pdffont.PDFFont = PDFFont
#pdfminer.pdffont.PDFSimpleFont = PDFSimpleFont
#pdfminer.pdffont.PDFType1Font = PDFType1Font
#pdfminer.pdffont.PDFTrueTypeFont = PDFTrueTypeFont
#pdfminer.pdffont.PDFType3Font = PDFType3Font
#pdfminer.pdffont.PDFCIDFont = PDFCIDFont
